/*  ----------------------------------------------------------------------------------
 *  triac_controller, provides zero detect function. on-off or timer application.
 *  with autostart or manual start of timer
 *  Triac_with_zd_32, 29.11.2019, by roland stiglmayr, follows v632
 *  system-clock = 11.0592 MHz. 
 *  uses Watchdog timer interrupt. ports adapted to new hardware. 
 *  ----------------------------------------------------------------------------------
 *  
 *  the function of this application is selected by hardware settings according 
 *  to the following table
 *  
 *  ON_OFF_EN   START_D     Function
 *  
 *  HIGH        HIGH/ LOW   On_Off_App, active while START_IN is LOW
 *  LOW         HIGH        Timer_App, timer starts @ START_IN = LOW
 *                          longer pressing of START_IN doubles time
 *  LOW         LOW         Auto_Timer, timer is started directly at PowerOn                        
 *
 *  START_D (start_direct) on ICSP pin4  
 *  timer timeout after MANUAL_TIME_SEC in seconds
 *  auto timeout after AUTO_TIME_SEC in seconds
 *  
 *  --------------------------------------------------------------------------------- */


#define TRIG_OUT_PIN 5                    //arduino style D5
#define LED_OUT_PIN 7                     //arduino style D7
#define START_PIN 9                       //arduino style D9, PB1
#define ON_OFF_EN_PIN 8                   //arduino style D8, PB0
#define START_D_PIN 11                    //arduino style D11, PB3
#define ZERO_DET_PIN 2                    //arduino style D2= INT0

//declare I/Os for using the fast macro commands
//
  #define TRIG_P PORTD                    //TRIG_OUT @PORTD 
  #define TRIG_OUT  PD5                   //TRIG_OUT @PD5, triac-gate
  #define LED_P PORTD                     //LED @PORTD
  #define LED_OUT PD7                     //LED @PD7, heartbeat LED
  #define START_IN_P PINB                 //Signal START @PINB
  #define START_IN PB1                    //low active signal on-off, start timer
  #define START_D_IN_P PINB               //signal START_D_IN @ PINB
  #define START_D_IN PB3                  //set type of start at timer
  #define ON_OFF_EN_P PINB                //signal ON_OFF_EN @ PINB
  #define ON_OFF_EN PB0                   //signal selects function
  #define ZERO_DET_PORT PIND              //ZERO_DET @ PIND
  #define ZERO_DET_IN PD2                 //detects mains zero voltage, PD2= INT0


#define EN_INT0 01                        //enables INT0, written to EIMSK
#define INT0_CHANGE 01                    //INTO on logic change -> EICRA
#define INT_CLEAR 03                      //clears pending INT0, INT1 -> EIFRA

#define T1_STOP 00                        //disable counting
#define T1_START 02                       //T1-CK, Tcycle*8= 0.72µs (TCCR1B)
#define T1_TRIGTIME (0xffff-8)            //triggerpulse length 8*.72µs + 4µs = 10µs

#define T2_STOP 00                        //disables counting
#define T2_START 07                       //T2-CK, Tcycle*1024= 92.593µs (TCCR2B)
#define T_ONE_HW 107                      //(107+1) * 92.6µs= 10ms
#define PHASE_MAX  100                    //fill pulses until (100+1) * 92.6µs= 9.4ms
#define PHASE_ANGLE_FRST 2                //length of first pulse 185µs, fill pulses follow

#define MAX_CYCLE_COUNT 50                //50 * 20ms= 1 sec
#define MANUAL_TIME_SEC (2*60)            //default shut off time @ manual start
#define AUTO_TIME_SEC (5*60)              //shut off @ auto mode start

boolean Second_HW;                        //set if second halfwave is active
boolean Second_HW_Rdy= false;             //period finished
boolean Shut_Off;                         //shut off triac, load
boolean Power_Save= false;                //start power save mode
boolean Start_Released= false;            //input start inactive
boolean Wd_Rdy= false;
boolean On_Off_App=  true;                //keeps type of activated app
boolean Auto_Timer_App= false;            //autostart if timer

int Cycle_Counter= MAX_CYCLE_COUNT;       //decremented each 20 ms
unsigned int Sec_Counter= 0;              //counts elapsed seconds
unsigned int Shutoff_Time;                //keeps shutoff time @ auto/man


void setup() 
{   
  //  after reset, start with sleep mode to stabilize VCC. prepare the watchdog 
  //  for wakeup
  //
        asm ("WDR");                      //clear wd-timer
        WDTCSR= 0x1c;                     //set WDCE, WDE to enable setting 
        WDTCSR= 0x1c;                     //wd-timer 0.25s      
        WDTCSR= 0x14;                     //clear WDE
        WDTCSR= (WDTCSR | 0x40);          //set WDIE, enable WD-interrupt
        interrupts();                     //enable interrupts      

        SMCR= 0x0f;                       //prepare xtd-standby
        asm ("SLEEP");                    //enter sleep mode
        SMCR= 0x00;                       //finish, not necessary           
        noInterrupts();                   //disable interrupts, after wakeup
        
        MCUSR= (MCUSR & 0xf7);            //clear WDRF, wd-reset active
        WDTCSR= WDTCSR | 0x18;            //set WDCE, enable setting WDE
        WDTCSR= 00;                       //clear WDIE, disables wd-timer
         
  
  digitalWrite (TRIG_OUT_PIN,HIGH);       //avoid glitch
  pinMode (TRIG_OUT_PIN,OUTPUT);          //output to triac-gate
  digitalWrite (TRIG_OUT_PIN,HIGH);       //disable triac
  
  pinMode (LED_OUT_PIN, OUTPUT);          //heartbeat LED
  pinMode (ZERO_DET_PIN, INPUT);          //input INT0, zerodetect
  pinMode (START_PIN, INPUT);             //start signal
  pinMode (ON_OFF_EN_PIN, INPUT);         //select sig on_off/ timer
  pinMode (START_D_PIN, INPUT_PULLUP);    //select sig start directly
    
  ACSR= 0x80;                             // disable analog comparator
                                          //to save power

  TIMSK0= 0;                              //disable timer ints
  TIMSK1= 0;
  TIMSK2= 0;
  
  // init INT0 -----------------------------------------------------
  //  
  EIMSK= 0;                               //disable hw ints
  PCICR= 0;                               //disable pinchange ints
  EICRA= INT0_CHANGE;                     //INT0 on level change
  EIFR= INT_CLEAR;                        //clear pending INT-flags


  //  init 16bit-timer 1, duration of trigger pulse ----------------
  //
  TCCR1A= 0x00;                           //normal operation
  TCCR1B= T1_STOP;                        //stop counter
  TCNT1= T1_TRIGTIME;                     //prepare first cycle
  OCR1AL= 0x00;
  OCR1BL= 0x00;
  TIMSK1= 0x01;                           //enable T1 overflow interrupt
  TIFR1= 0x07;                            //clear pending interrupt flags

  
  //  init 8bit-timer 2, generates second cycle ---------------------
  //
  TCCR2A= 0x02;                           //clear T2 on compareA match
  TCCR2B= T2_STOP;                        //stop T2
  TCNT2= 00;                              //clear counter T2
  OCR2A= T_ONE_HW;                        //one shot one halfwave
  OCR2B= PHASE_ANGLE_FRST;                //lowest phase at start
  TIMSK2= 0x06;                           //interrupt on A+B match
  TIFR2= 0x07;                            //clear interpt flags

  Second_HW= true;                        //allow synchron start with ZD  
  Cycle_Counter= MAX_CYCLE_COUNT;         //counts down at 1 second

  //  test the selected function as set by hardware
  //  ---------------------------------------------

  if (bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      bitRead (START_D_IN_P, START_D_IN))  //on_off_en, start_d_in = 11, ON-OFF
      On_Off_App= true;

  if (bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      !bitRead (START_D_IN_P, START_D_IN))  //on_off_en, start_d_in = 10, ON-OFF
      On_Off_App= true;

  if (!bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      !bitRead (START_D_IN_P, START_D_IN))  //on_off_en, start_d_in = 00, autotimer
      {
        On_Off_App= false;
        Auto_Timer_App= true;
      }

  if (!bitRead (ON_OFF_EN_P, ON_OFF_EN) &&
      bitRead (START_D_IN_P, START_D_IN))  //on_off_en, start_d_in = 01, manual timer
      {
        On_Off_App= false;
        Auto_Timer_App= false;
      }
      
  if (On_Off_App)                         //test if on-off app is selected
      Shut_Off= true;                     //disable triggerpulses until start
      
  if (!On_Off_App && !Auto_Timer_App)
      {
        Shut_Off= true;                   //no pulses until manual start
        Shutoff_Time= MANUAL_TIME_SEC;    //prepare manual shutoff time
        Start_Released= false;
      }
      
  if (!On_Off_App && Auto_Timer_App)
      {
        Shut_Off= false;                  //start immediately with pulses
        Shutoff_Time= AUTO_TIME_SEC;      //prepare manual shutoff time
      }          

  pinMode (START_D_PIN, INPUT);           //disable pullup, save power
 
  Power_Save= true;                       //save power until first cycle
  EIFR= INT_CLEAR;                        //clear pending hwints
  interrupts ();                          //now enable interrupts
  EIMSK= EN_INT0;                         //enable as first event the HW interrupt
}

  //-------------------------------------------------------------------
  //  WDT SERVICE ROUTINE, started after each reset, stretches the
  //  reset phase with low power consumption until power is stable
  //  only used once for wake up of sleep mode
  //-------------------------------------------------------------------
  
    ISR (WDT_vect)
        {
          Wd_Rdy=true;                    //flag not used
        }


  //---------------------------------------------------------------------
  //  HW INT0 INTERRUPT SERVICE ROUTINE, falling edge at INT0 starts ISR.
  //  following edge is disabled. INT0 driven by AC zero detector, ZD.
  //  OCR2B is set to length of first pulse. 
  //---------------------------------------------------------------------

  ISR (INT0_vect)
    { 
     if (Second_HW && !bitRead (ZERO_DET_PORT, ZERO_DET_IN))  //test if new period starts
                                                              //with low @ zero detect
        {         
          if (!Shut_Off)
              bitClear (TRIG_P, TRIG_OUT);          //insert longer trigger pulse              
          
          TCCR2B= T2_START;                         //synchron start of T2 generating T2A + T2B
          OCR2B= PHASE_ANGLE_FRST;                  //length of first pulse

          bitClear (LED_P, LED_OUT);                //LED off if active                     
          EIMSK= 00;                                //no more hw-trigger while this period
                                                    //protect against transients
          Second_HW= false;                         //first HW is started
          TIMSK2= 06;                               //enable compare match @ a+b
          --Cycle_Counter;                          //decrement each 20 ms
            if ((Cycle_Counter < 5) && (!Shut_Off)) //show heart beat @ eof second
                 bitSet (LED_P, LED_OUT);           //LED on                  
        }
     Power_Save= true;                              //reduce power until T2 or INT0 
    }

  //-------------------------------------------------------------------
  //  TIMER 2A INTERRUPT SERVICE ROUTINE, called 10 ms after INT0.
  //  starts second halfwave. COMPA-match clears TCNT2 and so 
  //  synchronizes COMPB to this event. 
  //  OCR2B is set to length of first pulse. 
  //-------------------------------------------------------------------

    ISR (TIMER2_COMPA_vect)
        {
          if (!Second_HW)                           //test if first HW is active
              {
                if (!Shut_Off)
                    bitClear (TRIG_P, TRIG_OUT);    //insert longer trigger pulse
                                  
                if ((Cycle_Counter < 5) && (!Shut_Off))  //show activation @ eof second
                      bitSet (LED_P, LED_OUT);      //LED on 
                                                      
                OCR2B= PHASE_ANGLE_FRST;            //length of first pulse
                Second_HW= true;                    //next event is second HW 
              }       

          Power_Save= true;                         //save power until T2B or INT0                 
        }


  //---------------------------------------------------------------------
  //  TIMER 2 COMPB ISR, started by INT0 or COMPB. starts a trigger pulse
  //  generates further pulses as long as OCR2B < PHASE_MAX by 
  //  preparing further T2B matches. tests on end of second halfwave
  //---------------------------------------------------------------------
  
  ISR (TIMER2_COMPB_vect)
      {
        if (!Shut_Off)                              //test if active
              bitClear (TRIG_P, TRIG_OUT);          //start trigger pulse
              
         TCNT1= T1_TRIGTIME;                        //prepare triggertime
         TCCR1B= T1_START;                          //start T1, enables power_save
                  
         bitClear (LED_P, LED_OUT);                 //LED off if active
          

         if (OCR2B <= (PHASE_MAX))                  //test if phasedelay < phase max         
            {
              OCR2B= (OCR2B + 4);                   //prepare next fill puls after 4*92.5µs
              if (OCR2B <= (PHASE_MAX/4))           //test if phasedelay < phase max
                  OCR2B= (OCR2B - 2);               //prepare next fill puls after 2*92.5µs                
            }
         else
            {                                       //no further 2B match follows
              if (Second_HW)
                {
                  EIFR= INT_CLEAR;                  //clear pending hw-interrupts  
                  EIMSK= EN_INT0;                   //enable INT0 again
                  TCCR2B= T2_STOP;                  //stop T2, no second T2A
                  TCNT2= 0;                         //synchronize T2 to zero-detect
                  Second_HW_Rdy= true;              //period elapsed
                }
            }
      }

  //-----------------------------------------------------------------------------
  //  TIMER 1 INTERRUPT SERVICE ROUTINE, finishes trigger pulse
  //-----------------------------------------------------------------------------

    ISR (TIMER1_OVF_vect)
        {
          bitSet (TRIG_P, TRIG_OUT);                //finish trigger pulse
          TCCR1B= T1_STOP;                          //stop T1
          Power_Save= true;                         //prepare power-save 
        }


  //---------------------------------------------------------------------------------
  //    main loop
  //    either on-off service function or timer service function is compiled.
  //    power save/ idle mode is started. sleeps until wake up event (INT0, T2A, T2B)
  //---------------------------------------------------------------------------------

  void loop() 
    {
      if (Power_Save)
       {
        if (Shut_Off == true)
            Cycle_Counter= MAX_CYCLE_COUNT;         //prepare next activation
        
        
        if (Cycle_Counter <= 0)                     //test if one second is finished
          {
            Cycle_Counter= MAX_CYCLE_COUNT;         //preset cycle_counter
             if (!Shut_Off)                      
                 ++Sec_Counter;                     //if not yet ready, inc second            
          }

      //  procedure for timer application
      //  -------------------------------

      if (!On_Off_App)                              //test on timer_app
        {
          if (Second_HW_Rdy)                        //test if end of period
             {
               Second_HW_Rdy= false;                //finish this period

              if (Shut_Off && bitRead (START_IN_P, START_IN)) //test if start is released
                  Start_Released= true;

              if (Sec_Counter >= Shutoff_Time)    //test if timer is ready
                  {
                    Shut_Off= true;                 //disable trigger pulses
                    Sec_Counter= 0;                 //prepare restart
                    Start_Released= false;          //key must be released                    
                  }
                
            //  test on start/ restart condition  -------------------------
            //  no retrigger if auto-start        ------
            
              if (Auto_Timer_App)
                  {            
                    if (Auto_Timer_App && !bitRead (START_IN_P, START_IN) && 
                    Shut_Off && Start_Released && (Sec_Counter == 0)) 

                      {
                        Shut_Off= false;            //restart
                        Shutoff_Time= AUTO_TIME_SEC;  //prepare shutoff_time
                        Start_Released= false;      //start must be released
                      }
                  }
            
              //start-restart if manual-start      
              //short activation > manual-time, longer activation > double manual-time
            
              if (!Auto_Timer_App && !bitRead (START_IN_P, START_IN) &&
                  Start_Released)
                  {
                    if (Sec_Counter == 0)
                      {
                        Shut_Off= false;
                        Shutoff_Time= MANUAL_TIME_SEC;      //prepare shutoff_time
                      }
                    if (Sec_Counter == 1)
                        Shutoff_Time= MANUAL_TIME_SEC * 2;  //time * 2 if active longer than 1 second

                    if (Sec_Counter == 2)
                        Shutoff_Time= MANUAL_TIME_SEC * 4;   //time * 4 if active longer than 3 seconds
                  }
           }
        }       
                    
        // procedure for ON-OFF application
        // ---------------------------------

          if (On_Off_App)
            {
              if (Second_HW_Rdy)                    //test eof period, then service
                 {
                  Sec_Counter= 0;                   //no second counter
                  Second_HW_Rdy= false;

                  if (!bitRead (START_IN_P, START_IN))//test start signal    
                      Shut_Off= false;              //enable trigger
                  else
                      Shut_Off= true;               //disable trigger
                 }
            }         

      Power_Save= false;        //disables power save after wake up
      SMCR= 0x0f;               //prepare power save
      asm ("SLEEP");
      SMCR= 0x00;               //finish sleep
    }
}


